#include "stdafx.h"
#include "ArchiveManager.h"
#include "ArchiveExtractor.h"
#include "Error.h"

#pragma comment(lib, "Version.lib")

///////////////////////////////////////
// Utility Macro Functions
#define IF_INVALID_HANDLE(h) \
	if (h == NULL) { AM_SetError(ERROR_INVALID_HANDLE); } \
	if (h == NULL)

#define HANDLE_TO_EXTRACTOR(h) \
	(static_cast<CArchiveExtractor*>(h))

#define SIZEOFARRAY(x) (sizeof(x)/sizeof(*x))

///////////////////////////////////////
// Global variables
int g_dwTlsIndex = TLS_OUT_OF_INDEXES;
HINSTANCE g_hInstance = NULL;

///////////////////////////////////////
// Static global variables
static CMyComPtr<ICompressCodecsInfo> g_compressCodecsInfo;
static CCodecs *g_codecs;
static int g_TypeToIndex[6] = { -1, };
static WCHAR *g_MainDll = L"7z.dll";

// Supported archive sorted by ARCHIVE_???
static const WCHAR *g_supportedArchive[] = {
	L"Zip", // 1 ...
	L"Lzh",
	L"Cab", 
	L"Rar",
	L"Iso",
};

///////////////////////////////////////
// Functions
BOOL InitProcess(HMODULE hModule);
void FilterArchive();
void FinalizeProcess();
void NewTLS();
void DeleteTLS();

///////////////////////////////////////
// Implemention

BOOL APIENTRY DllMain( HMODULE hModule,
                       DWORD  ul_reason_for_call,
                       LPVOID lpReserved
					 )
try
{
	switch (ul_reason_for_call)
	{
	case DLL_PROCESS_ATTACH:
		return InitProcess(hModule);
	case DLL_THREAD_ATTACH:
		NewTLS();
		break;
	case DLL_THREAD_DETACH:
		DeleteTLS();
		break;
	case DLL_PROCESS_DETACH:
		FinalizeProcess();
		break;
	}

    return TRUE;
} catch (...) {
	return FALSE;
}

BOOL InitProcess(HMODULE hModule)
{
	// Set Memoryleak for debuggin
	// vOIɃ[N`FbNAfobOo͂ɕ\B
	// _DEBUG `ĂȂƂ́AȂB
	_CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF);

	// this DLL handle
	g_hInstance = hModule;

	// Allocate TLS
	g_dwTlsIndex = TlsAlloc();

	// Failed
	if (g_dwTlsIndex == TLS_OUT_OF_INDEXES) {
		return FALSE;
	}

	// Create TLS for error code
	NewTLS();

	// Load Archive DLL
	g_codecs = new CCodecs();
	g_compressCodecsInfo = g_codecs;

    WCHAR dllname[MAX_PATH+1];
    if (GetModuleFileName(hModule, dllname, _countof(dllname)-1) == 0) {
        return FALSE;
    }
    WCHAR *wk = wcsrchr(dllname, L'\\');
    if (wk == NULL || *(wk+1) == L'\0') {
        return FALSE;
    }
    *(wk+1) = L'\0';
    wcsncat_s(dllname, _countof(dllname), g_MainDll, _TRUNCATE);

	//if (g_codecs->LoadDll(g_MainDll) != S_OK) {
	if (g_codecs->LoadDll(dllname) != S_OK) {
		return FALSE;
	}

	FilterArchive();

	return TRUE;
}

int IsSupported(const WCHAR *type)
{
	for (int i = 0; i < SIZEOFARRAY(g_supportedArchive); i++) {
		if (lstrcmpiW(type, g_supportedArchive[i]) == 0) {
			return i+1;
		}
	}
	return -1;
}

void FilterArchive()
{
	for (int i = 0; i < g_codecs->Formats.Size(); ) {
		int index = IsSupported(g_codecs->Formats[i].Name);
		if (index != -1) {
			g_TypeToIndex[index] = i;
			i++;
		} else {
			g_codecs->Formats.Delete(i);
		}
	}
}

void FinalizeProcess()
{
	// TLS for error code
	DeleteTLS();

	// Free TLS
	TlsFree(g_dwTlsIndex);
}

void NewTLS()
{
	// New ErrorValue in TLS
	TlsSetValue(g_dwTlsIndex, new ErrorValue());
}

void DeleteTLS()
{
	// Destroy ErrorValue in TLS
	delete static_cast<ErrorValue*>(TlsGetValue(g_dwTlsIndex));
}


ARCHIVEMANAGER_API DWORD WINAPI GetAMLastError()
{
	return AM_GetError();
}


ARCHIVEMANAGER_API DWORD WINAPI GetAMVersion()
{
	return ARCHIVE_MANAGER_VERSION;
}

ARCHIVEMANAGER_API BOOL WINAPI Get7zVersion(WCHAR *version, DWORD size)
{
	try {
		if (version == NULL) {
			return FALSE;
		}

		// Get FileVersionInfo size
		DWORD dwVerInfoSize = GetFileVersionInfoSize(g_MainDll, 0);
		if (dwVerInfoSize == 0) {
			return FALSE;
		}

		// new Memory and Get FileVersionInfo
		BYTE *pBlock = new BYTE[dwVerInfoSize];
		if (GetFileVersionInfo(g_MainDll, 0, dwVerInfoSize, pBlock) == 0) {
			goto error;
		}

		// Get codepage
		UINT cbTranslate;
		struct LANGANDCODEPAGE {
			WORD wLanguage;
			WORD wCodePage;
		} *lpTranslate;
		if (VerQueryValue(pBlock, L"\\VarFileInfo\\Translation", (LPVOID*)&lpTranslate, &cbTranslate) == 0) {
			goto error;
		}

		for (UINT i = 0; i < cbTranslate/sizeof(*lpTranslate); i++) {
			WCHAR name[256];
			wsprintf(name, L"\\StringFileInfo\\%04x%04x\\ProductVersion", lpTranslate[i].wLanguage,
				lpTranslate[i].wCodePage);
			LPVOID pVersion;
			UINT dwBytes;
			if (VerQueryValue(pBlock, name, (LPVOID*)&pVersion, &dwBytes)) {
				lstrcpynW(version, (LPCWSTR)pVersion, size);
				delete [] pBlock;
				return TRUE;
			}
		}

error:
		delete [] pBlock;
		return FALSE;
	} catch (...) {
		return FALSE;
	}
}

ARCHIVEMANAGER_API HARCHIVE WINAPI OpenArchive(const WCHAR *filename)
{
	try {
		// Create extractor
		CArchiveExtractor *archive = new CArchiveExtractor();
		if (!archive->Open(g_codecs, filename)) {
			delete archive;
			return NULL;
		}
		
		AM_SetError(ERROR_SUCCESS);
		return static_cast<HARCHIVE>(archive);
	} catch (std::bad_alloc) {
		AM_SetError(ERROR_OUTOFMEMORY);
		return NULL;
	} catch (...) {
		AM_SetError(ERROR_INTERNAL_ERROR);
		return NULL;
	}
}


ARCHIVEMANAGER_API HARCHIVE WINAPI OpenArchiveAs(const WCHAR *filename, ArchiveType type)
{
	try {
		int index = -1;
		if (type != 0 && type <= ARCHIVE_ISO) {
			index = g_TypeToIndex[type];
		}

		// Is supported
		if (index == -1) {
			AM_SetError(ERROR_UNSUPPORTED_TYPE);
			return FALSE;
		}

		// Create extractor
		CArchiveExtractor *archive = new CArchiveExtractor();
		if (!archive->OpenAs(g_codecs, filename, index)) {
			delete archive;
			return NULL;
		}
		
		AM_SetError(ERROR_SUCCESS);
		return static_cast<HARCHIVE>(archive);
	} catch (std::bad_alloc) {
		AM_SetError(ERROR_OUTOFMEMORY);
		return NULL;
	} catch (...) {
		AM_SetError(ERROR_INTERNAL_ERROR);
		return NULL;
	}
}


ARCHIVEMANAGER_API HARCHIVE WINAPI OpenArchiveFromHandle(HANDLE hFile)
{
	try {
		// Create extractor
		CArchiveExtractor *archive = new CArchiveExtractor();
		if (!archive->Open(g_codecs, hFile)) {
			delete archive;
			return NULL;
		}
		
		AM_SetError(ERROR_SUCCESS);
		return static_cast<HARCHIVE>(archive);
	} catch (std::bad_alloc) {
		AM_SetError(ERROR_OUTOFMEMORY);
		return NULL;
	} catch (...) {
		AM_SetError(ERROR_INTERNAL_ERROR);
		return NULL;
	}
	return NULL;
}

ARCHIVEMANAGER_API HARCHIVE WINAPI OpenArchiveAsFromHandle(HANDLE hFile, ArchiveType type)
{
	try {
		int index = -1;
		if (type != 0 && type <= ARCHIVE_ISO) {
			index = g_TypeToIndex[type];
		}

		// Is supported
		if (index == -1) {
			AM_SetError(ERROR_UNSUPPORTED_TYPE);
			return FALSE;
		}

		// Create extractor
		CArchiveExtractor *archive = new CArchiveExtractor();
		if (!archive->OpenAs(g_codecs, hFile, index)) {
			delete archive;
			return NULL;
		}
		
		AM_SetError(ERROR_SUCCESS);
		return static_cast<HARCHIVE>(archive);
	} catch (std::bad_alloc) {
		AM_SetError(ERROR_OUTOFMEMORY);
		return NULL;
	} catch (...) {
		AM_SetError(ERROR_INTERNAL_ERROR);
		return NULL;
	}
	return NULL;
}

ARCHIVEMANAGER_API BOOL WINAPI CloseArchive(HARCHIVE hArchive)
{
	try {
		if (hArchive != NULL) {
			// Destroy extractor
			delete HANDLE_TO_EXTRACTOR(hArchive);
		}
		AM_SetError(ERROR_SUCCESS);
		return TRUE;
	} catch (...) {
		AM_SetError(ERROR_INTERNAL_ERROR);
		return FALSE;
	}
}


ARCHIVEMANAGER_API ArchiveType WINAPI GetArchiveType(HARCHIVE hArchive)
{
	try {
		IF_INVALID_HANDLE(hArchive) {
			return ARCHIVE_NONE;
		}
		
		// Search from Index
		int index = HANDLE_TO_EXTRACTOR(hArchive)->GetArchiveIndex();
		for (int i = 1; i < SIZEOFARRAY(g_TypeToIndex); i++) {
			if (g_TypeToIndex[i] == index) {
				AM_SetError(ERROR_SUCCESS);
				return i;
			}
		}
		
		// Cannot be through
		_ASSERTE(FALSE);
		AM_SetError(ERROR_INTERNAL_ERROR);
		return ARCHIVE_NONE;
	} catch (...) {
		AM_SetError(ERROR_INTERNAL_ERROR);
		return ARCHIVE_NONE;
	}
}

ARCHIVEMANAGER_API UInt32 GetFileCount(HARCHIVE hArchive)
{
	try {
		IF_INVALID_HANDLE(hArchive) {
			return 0;
		}
		return HANDLE_TO_EXTRACTOR(hArchive)->GetFileList().size();
	} catch (...) {
		AM_SetError(ERROR_INTERNAL_ERROR);
		return 0;
	}
}

ARCHIVEMANAGER_API UInt64 GetFileSizeFromArchive(HARCHIVE hArchive, UInt32 index)
{
	try {
		IF_INVALID_HANDLE(hArchive) {
			return 0;
		}
		// Get extractor and file list
		CArchiveExtractor *extractor = HANDLE_TO_EXTRACTOR(hArchive);
		ConstFileList list = extractor->GetFileList();

		UInt64 size = list[index]->Size;

		AM_SetError(ERROR_SUCCESS);
		return size;
	} catch (std::out_of_range) {
		AM_SetError(ERROR_INVALID_INDEX);
		return 0;
	} catch (...) {
		AM_SetError(ERROR_INTERNAL_ERROR);
		return 0;
	}
}

ARCHIVEMANAGER_API DWORD GetFileNameFromArchive(HARCHIVE hArchive, UInt32 index, WCHAR *str, DWORD strSize)
{
	try {
		IF_INVALID_HANDLE(hArchive) {
			return 0;
		}

		// Get extractor and file list
		CArchiveExtractor *extractor = HANDLE_TO_EXTRACTOR(hArchive);
		ConstFileList list = extractor->GetFileList();

		// Check pointer and buffer size
		DWORD len = list[index]->Name.Length() + 1;
		if (str == NULL || strSize == 0 || strSize < len) {
			return len;
		}

		// Copying
		lstrcpynW(str, list[index]->Name, strSize);

		AM_SetError(ERROR_SUCCESS);
		return len;
	} catch (std::out_of_range) {
		AM_SetError(ERROR_INVALID_INDEX);
		return 0;
	} catch (...) {
		AM_SetError(ERROR_INTERNAL_ERROR);
		return 0;
	}
}

ARCHIVEMANAGER_API DWORD GetFilePathFromArchive(HARCHIVE hArchive, UInt32 index, WCHAR *str, DWORD strSize)
{
	try {
		IF_INVALID_HANDLE(hArchive) {
			return 0;
		}

		// Get extractor and file list
		CArchiveExtractor *extractor = HANDLE_TO_EXTRACTOR(hArchive);
		ConstFileList list = extractor->GetFileList();

		// Check pointer and buffer size
		DWORD len = list[index]->Name.Length() + 1;
		if (str == NULL || strSize == 0 || strSize < len) {
			return len;
		}

		// Copying
		lstrcpynW(str, list[index]->Path, strSize);

		AM_SetError(ERROR_SUCCESS);
		return len;
	} catch (std::out_of_range) {
		AM_SetError(ERROR_INVALID_INDEX);
		return 0;
	} catch (...) {
		AM_SetError(ERROR_INTERNAL_ERROR);
		return 0;
	}
}

ARCHIVEMANAGER_API BOOL ExtractToFile(HARCHIVE hArchive, UInt32 index, const WCHAR *path, ArchiveExtractCallback func, LPVOID param)
{
	try {
		IF_INVALID_HANDLE(hArchive) {
			return FALSE;
		}
		return HANDLE_TO_EXTRACTOR(hArchive)->ExtractToFile(index, path, func, param) ? TRUE : FALSE;
	} catch (...) {
		AM_SetError(ERROR_INTERNAL_ERROR);
		return FALSE;
	}
}

ARCHIVEMANAGER_API BOOL ExtractToMemory(HARCHIVE hArchive, UInt32 index, BYTE *buff, UInt32 buffSize, ArchiveExtractCallback func, LPVOID param)
{
	try {
		IF_INVALID_HANDLE(hArchive) {
			return FALSE;
		}
		return HANDLE_TO_EXTRACTOR(hArchive)->ExtractToMemory(index, buff, buffSize, func, param) ? TRUE : FALSE;
	} catch (...) {
		AM_SetError(ERROR_INTERNAL_ERROR);
		return FALSE;
	}
}


ARCHIVEMANAGER_API BOOL ExtractToHandle(HARCHIVE hArchive, UInt32 index, HANDLE handle, ArchiveExtractCallback func, LPVOID param)
{
	try {
		IF_INVALID_HANDLE(hArchive) {
			return FALSE;
		}
		return HANDLE_TO_EXTRACTOR(hArchive)->ExtractToHandle(index, handle, func, param) ? TRUE : FALSE;
	} catch (...) {
		AM_SetError(ERROR_INTERNAL_ERROR);
		return FALSE;
	}
}